/*
    (C) 1995-2001 AROS - The Amiga Research OS
    (C) 2003-2018 MorphOS Team
    $Id: delete.c,v 1.3 2018/08/13 02:26:31 piru Exp $
 
    Desc: Delete CLI Command
    Lang: English
*/

#define USE_SANELINKMODE      1
#define USE_DELAYEDDELETE     0
#define USE_CBMPROTERRCOMPAT  1

#define	DEBUG_GEN(x)		;


/*****************************************************************************
 
    NAME
 
    Delete
 
    SYNOPSIS
 
    Delete { (name | pattern) } [ALL] [QUIET] [FORCE] [FOLLOWLINKS]
 
    LOCATION
 
	Workbench/c
 
    FUNCTION
 
    Deletes files and directories. You may delete several files and directories
    by listing them separately or by using wildcards. To abort a multiple
    delete, press CTRL-C. Delete will notify the user of which files it
    weren't able to delete.
        Delete cannot delete directories which are not empty unless the
    ALL option is used. To suppresss file and directory names from being
    printed while deleted use the QUIET option. If the 'd' protection bit
    is cleared for a file or directory, it may not be deleted unless the
    FORCE option is used.
 
    INPUTS
 
    FILE/M/A  --  files or directories to delete (may contain patterns)
    ALL/S     --  recursively delete dirctories
    QUIET/S   --  don't print which files/directories were deleted
    FORCE/S   --  delete files/directories even if they are protected from
                  deletion
    FOLLOWLINKS/S
              --  follow hard/soft directory links when deleting. default
                  do NOT follow links. this is different to original delete
                  which always follow linked directories.
 
    RESULT
 
    NOTES
 
    EXAMPLE
 
    Delete RAM:T/#? ALL FORCE
 
    Deletes all directories and files recursively in the directory RAM:T
    even if they are protected from deletion.
 
 
    BUGS
 
    SEE ALSO
 
    INTERNALS
 
    Shows a good usage of the pattern matching capabilities.
 
    HISTORY
 
    16.11.2002  Piru    --  Print "No file to delete" if no files were deleted
    18.07.2002  Piru    --  Added FOLLOWLINKS/S, default don't follow links!
    25.11.2000  SDuvan  --  (Re)Implemented (the old version didn't no much)
 
******************************************************************************/

#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dosasl.h>
#include <dos/exall.h>

#include <proto/dos.h>
#include <proto/exec.h>

#include <string.h>


#if USE_SANELINKMODE

#define ARG_TEMPLATE	"FILE/M/A,ALL/S,QUIET/S,FORCE/S,FOLLOWLINKS/S"

#else  /* USE_SANELINKMODE */

#define ARG_TEMPLATE	"FILE/M/A,ALL/S,QUIET/S,FORCE/S"

#endif /* USE_SANELINKMODE */


enum
{
    ARG_FILE = 0,
    ARG_ALL,
    ARG_QUIET,
    ARG_FORCE,

#if USE_SANELINKMODE

    ARG_FOLLOWLINKS,

#endif /* USE_SANELINKMODE */

    NOOFARGS
};


/* Maximum file path length */
#define MAX_PATH_LEN    2048


#include "delete_version.h"

static const char version[] = VERSTAG;


#if USE_SANELINKMODE

static int doDelete(struct AnchorPath *ap, STRPTR *files, BOOL all, BOOL quiet,
                    BOOL force, BOOL forcelinks, struct DosLibrary *DOSBase);

#else /* USE_SANELINKMODE */

static int doDelete(struct AnchorPath *ap, STRPTR *files, BOOL all, BOOL quiet,
                    BOOL force, struct DosLibrary *DOSBase);

#endif /* USE_SANELINKMODE */


int main(void)
{
    struct ExecBase    *SysBase;
    struct DosLibrary  *DOSBase;
    struct RDArgs      *rda;
    struct AnchorPath  *ap;
    IPTR                args[NOOFARGS] = { 0,
                                           FALSE,
                                           FALSE,
                                           FALSE
#if USE_SANELINKMODE
                                           ,FALSE
#endif /* USE_SANELINKMODE */
                                         };
    int	 retval = RETURN_OK;

    SysBase=*((struct ExecBase**) 4);
    if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37))!=0)
    {
        ap = AllocVec(sizeof(struct AnchorPath) + MAX_PATH_LEN,
                      MEMF_ANY | MEMF_CLEAR);

        if (ap)
        {
            ap->ap_Strlen = MAX_PATH_LEN;

            rda = ReadArgs(ARG_TEMPLATE, args, NULL);

            if (rda)
            {
                /* Convert arguments into (less complex) variables */
                STRPTR *files = (STRPTR *)args[ARG_FILE];
                BOOL    all = (BOOL)args[ARG_ALL];
                BOOL    quiet = (BOOL)args[ARG_QUIET];
                BOOL    force = (BOOL)args[ARG_FORCE];
#if USE_SANELINKMODE
                BOOL    followlinks = (BOOL)args[ARG_FOLLOWLINKS];

		if (DOSBase->dl_lib.lib_Version > 50 ||
		    (DOSBase->dl_lib.lib_Version == 50 &&
		     DOSBase->dl_lib.lib_Revision >= 67))
		{
                    ap->ap_Strlen   |= APSF_EXTENDED;
                    ap->ap_Extended  = APEF_LiteralSLinks;

                    if (followlinks)
                    {
                        #warning "APF_FollowHLinks should not be set?"
                        ap->ap_Flags |= APF_FollowHLinks;
                    }
                    else
                    {
                        ap->ap_Extended |= APEF_DontFollowSLinks;
                    }
                }

                retval = doDelete(ap, files, all, quiet, force, followlinks, DOSBase);

#else /* USE_SANELINKMODE */

                retval = doDelete(ap, files, all, quiet, force, DOSBase);

#endif /* USE_SANELINKMODE */

                FreeArgs(rda);
            }
            else
            {
                PrintFault(IoErr(), "Delete");
                retval = RETURN_WARN;
            }
        }
        else
        {
            retval = RETURN_FAIL;
        }

        FreeVec(ap);
        CloseLibrary((struct Library *)DOSBase);
    }
    return retval;
} /* main */


#if USE_SANELINKMODE

static inline
BOOL isDirectory(struct AnchorPath *ap, BOOL followflag, struct DosLibrary *DOSBase)
{
    BOOL isdir;

    DEBUG_GEN(Printf("isDirectory \"%s\" - entry\n", ap->ap_Buf));

    if (ap->ap_Info.fib_DirEntryType == ST_SOFTLINK)
    {
        /* Softlink to dir/file - Default don't enter it, unless
           flag is set and the destination is really softlink.

           We have all lost some files due to braindead behaviour
           of the original delete regarding softlinks. - Piru
        */
        isdir = FALSE;

        if (followflag)
        {
            /* Okay flag set, figure out if this is a softlink to directory
            */
            BPTR lock;

            lock = Lock(ap->ap_Buf, ACCESS_READ);
            if (lock)
            {
                struct FileInfoBlock *fib;

                fib = AllocDosObject(DOS_FIB, NULL);
                if (fib)
                {
                    if (Examine(lock, fib))
                    {
                        /* Just extra sanity check so it can't be softlink
                           anymore (weird, fucked up, fs?).
                        */
                        isdir = (fib->fib_DirEntryType >= 0 &&
                                 fib->fib_DirEntryType != ST_SOFTLINK);
                    }
                    DEBUG_GEN(
                        else
                        {
                            Printf("Examine failed for \"%s\"\n", ap->ap_Buf);
                        }
                    );

                    FreeDosObject(DOS_FIB, fib);
                }
                DEBUG_GEN(else
                          {
                              Printf("AllocDosObject failed\n");
                          }
                         );

                UnLock(lock);
            }
            DEBUG_GEN(else
                      {
                          Printf("Lock failed for \"%s\"\n", ap->ap_Buf);
                      }
                     );
        }
    }
    else if (ap->ap_Info.fib_DirEntryType == ST_LINKDIR)
    {
        /* Hardlink to directory - Only follow it if flag set.

           It is debatable whether hardlinks should be followed by
           default. IMHO not. - Piru
        */
        isdir = followflag;
    }
    else
    {
        isdir = ap->ap_Info.fib_DirEntryType >= 0;
    }

    DEBUG_GEN(Printf("isDirectory \"%s\", return %s\n", ap->ap_Buf, isdir ? "TRUE" : "FALSE"));
    return isdir;
}

#else /* USE_SANELINKMODE */

#define isDirectory(ap,f,dosbase) ((ap)->ap_Info.fib_DirEntryType >= 0)

#endif /* USE_SANELINKMODE */

#define isDeletable(fib) (!((fib)->fib_Protection & FIBF_DELETE))


#if USE_SANELINKMODE

static int doDelete(struct AnchorPath *ap, STRPTR *files, BOOL all, BOOL quiet,
                    BOOL force, BOOL forcelinks, struct DosLibrary *DOSBase)

#else /* USE_SANELINKMODE */

static int doDelete(struct AnchorPath *ap, STRPTR *files, BOOL all, BOOL quiet,
                    BOOL force, struct DosLibrary *DOSBase)

#endif /* USE_SANELINKMODE */

{
    LONG  match = 0;
    int   i;
#if USE_DELAYEDDELETE
    char  name[MAX_PATH_LEN];
    BOOL  isfile = TRUE;
    BOOL  deleteit = FALSE;
#endif /* USE_DELAYEDDELETE */
    BOOL  deletedfile = FALSE;
    BOOL  firstmatch = TRUE;
    STRPTR file;
#if USE_CBMPROTERRCOMPAT
    LONG  rc = RETURN_OK;
    LONG  lastioerr = 0;
#endif /* USE_CBMPROTERRCOMPAT */

    if (!files)
    {
        return RETURN_OK;
    }

    for (i = 0; (file = files[i]); i++)
    {
        /* Check if someone tries to delete a device/volume root.
         * This is not allowed. - Piru
         */
#if 1
        struct DevProc *dp;

        dp = GetDeviceProc(file, NULL);
        if (dp)
        {
            int len;

            FreeDeviceProc(dp);

            len = strlen(file);
            if (len && file[len - 1] == ':')
            {
                Printf("%s is a device and cannot be deleted\n", (ULONG)file);
                return RETURN_WARN;
            }
        }
#endif

        ap->ap_BreakBits  = SIGBREAKF_CTRL_C;
        ap->ap_FoundBreak = 0;

        for (match = MatchFirst(file, ap);
             match == 0;
             match = MatchNext(ap))
        {
            firstmatch = FALSE;
#if USE_CBMPROTERRCOMPAT
            lastioerr = 0;
#endif /* USE_CBMPROTERRCOMPAT */

#if USE_DELAYEDDELETE

            if (deleteit)
            {
                /* Try to delete the file or directory */
                DEBUG_GEN(Printf("DeleteFile \"%s\"\n", name));
                if (!DeleteFile(name))
                {
                    LONG ioerr = IoErr();
                    Printf("%s  Not Deleted", (ULONG)name);
                    PrintFault(ioerr, "");

#if 0
                    /* If ALL is given as a parameter, we continue */
                    if (ioerr != ERROR_DELETE_PROTECTED && !all && isfile)
                    {
                        MatchEnd(ap);

                        return RETURN_FAIL;
                    }
#endif
                }
                else if (!quiet)
                {
                    Printf("%s  Deleted\n", (ULONG)name);
                }

                deletedfile = TRUE;
                deleteit = FALSE;
            }

#endif /* USE_DELAYEDDELETE */

            /* If this is a directory, we enter it regardless if the ALL
               switch is set. */
            if (isDirectory(ap, forcelinks, DOSBase))
            {
                /* This is a directory. It's important to check if we just left
                   the directory or is about to enter it. This is because we
                   cannot delete a directory until we have deleted all the
                   files in it. */
                if (ap->ap_Flags & APF_DIDDIR)
                {
                    /* If we get here, we are in ALL mode and have deleted
                       all the files inside the dir (if none were protected
                       and such). */

                    ap->ap_Flags &= ~APF_DIDDIR;

                    /* Now go on and delete the directory */
                }
                else
                {
                    /* We stumbled upon a directory */

                    if (all)
                    {
                        ap->ap_Flags |= APF_DODIR;

                        /* We don't want to delete a directory before deleting
                           possible files inside it. Thus, match next file */
                        continue;
                    }

                    /* If all is not set, DeleteFile() will return an error
                       in case the directory was not empty. */
                }
            }

            /* Check permissions */

#if USE_DELAYEDDELETE
            /* Mark the entry for deletion */
            deleteit = TRUE;
#endif /* USE_DELAYEDDELETE */

            if (!isDeletable(&ap->ap_Info))
            {
                /* Consider delete protected file/dir 'deleted' */
                deletedfile = TRUE;

                if (force)
                    SetProtection(ap->ap_Buf, 0);
                else
                {
                    Printf("%s  Not Deleted", (ULONG)ap->ap_Buf);
                    PrintFault(ERROR_DELETE_PROTECTED, "");

#if USE_CBMPROTERRCOMPAT
                    /* Compatibility with C= delete, return error code */
                    rc = RETURN_FAIL;
                    lastioerr = ERROR_DELETE_PROTECTED;
#endif /* USE_CBMPROTERRCOMPAT */

                    /* 13-Jul-2003 bugfix: Don't stop but just skip file/dir - Piru */
#if USE_DELAYEDDELETE
                    deleteit = FALSE;
#else
                    continue;
#endif
                }
            }

#if USE_DELAYEDDELETE

            isfile = !isDirectory(ap, forcelinks, DOSBase);
            strcpy(name, ap->ap_Buf);

#else /* USE_DELAYEDDELETE */

            /* Try to delete the file or directory */
            DEBUG_GEN(Printf("DeleteFile \"%s\"\n", ap->ap_Buf));
            if (!DeleteFile(ap->ap_Buf))
            {
                LONG ioerr = IoErr();
                Printf("%s  Not Deleted", (ULONG)ap->ap_Buf);
                PrintFault(ioerr, "");

#if USE_CBMPROTERRCOMPAT
                /* Continue operation but indicate a failure eventually */
                lastioerr = ioerr;
                rc = RETURN_FAIL;
#endif /* USE_CBMPROTERRCOMPAT */

#if 0
                /* If ALL is given as a parameter, we continue */
                if (ioerr != ERROR_DELETE_PROTECTED && !all && !isDirectory(ap, forcelinks, DOSBase))
                {
                    MatchEnd(ap);

                    return RETURN_FAIL;
                }
#endif
            }
            else if (!quiet)
            {
                Printf("%s  Deleted\n", (ULONG)ap->ap_Buf);
            }

            deletedfile = TRUE;

#endif /* USE_DELAYEDDELETE */
        }

        if (ap->ap_FoundBreak & SIGBREAKF_CTRL_C)
        {
            MatchEnd(ap);

            PrintFault(ERROR_BREAK, NULL);
            return RETURN_ERROR;
        }

        MatchEnd(ap);

#if 0
        /* Softlink with dangling destionation fix... Not needed
           anymore, since this stuff is in MatchFirst/Next now.. - Piru */

        if (match == ERROR_OBJECT_NOT_FOUND)
        {
            struct DevProc *dvp = NULL;

            DEBUG_GEN(Printf("\"%s\" match %ld\n", file, match));

            dvp = GetDeviceProc(file, dvp);
            if (dvp)
            {
                BPTR lock;
                lock = CurrentDir(dvp->dvp_Lock);
                if (ReadLink(dvp->dvp_Port, dvp->dvp_Lock, file, name, sizeof(name)) >= 0)
                {
                    DEBUG_GEN(Printf("-> \"%s\"\n", name));
                }
                (void) CurrentDir(lock);
                FreeDeviceProc(dvp);
            }
        }
#endif /* 0 */

#if USE_DELAYEDDELETE

        if (deleteit)
        {
            /* Try to delete the file or directory */
            if (!DeleteFile(name))
            {
                LONG ioerr = IoErr();
                Printf("%s  Not Deleted", (ULONG)name);
                PrintFault(ioerr, "");

                /* If ALL is given as a parameter, we continue */
                //if (!all)
                //{
                //    return RETURN_FAIL;
                //}
            }
            else if (!quiet)
            {
                Printf("%s  Deleted\n", (ULONG)name);
            }

            deletedfile = TRUE;
            deleteit = FALSE;
        }

#endif /* USE_DELAYEDDELETE */

    }

    if (firstmatch && match &&
        match != ERROR_OBJECT_NOT_FOUND &&
        match != ERROR_NO_MORE_ENTRIES)
    {
        PrintFault(match, NULL);
        return RETURN_WARN;
    }

    if (!deletedfile)
    {
        PutStr("No file to delete\n");
        return RETURN_WARN;
    }

#if USE_CBMPROTERRCOMPAT
    if (match == ERROR_NO_MORE_ENTRIES)
    {
        SetIoErr(rc && lastioerr ? lastioerr : 0);
    }

    return rc;
#else
    if (match == ERROR_NO_MORE_ENTRIES)
    {
        SetIoErr(0);
    }

    return RETURN_OK;
#endif /* USE_CBMPROTERRCOMPAT */
}
